// Includes
#include "main.h"
#include "lcd.h"
#include <stdio.h>

// Private variables
SPI_HandleTypeDef hspi1;
UART_HandleTypeDef huart1;
UART_HandleTypeDef huart2;

// Private function prototypes
void SystemClock_Config(void);
static void MX_GPIO_Init(void);
static void MX_USART2_UART_Init(void);
static void MX_SPI1_Init(void);
static void MX_USART1_UART_Init(void);

// functions related to comm. with host PC via UART2;
void readFromHostPC( uint32_t timeout);

// functions related to BlueTooth (BT) module;
void readFromBT( uint32_t timeout, int debug);
void sendCommandToBT(char *command, uint32_t timeout, int debug);
void BT_module_rename(void);

// LCD related;
#define DEFAULT_FONT FONT_6X8

// we talk to the BT module using UART1;
# define BUFF_SIZE 512
char rx_buff_from_BlueTooth_module[BUFF_SIZE];
char tx_buff_to_BlueTooth_module[] = "Cris says hi to all of you!\r\n";

// at the same time we use UART2 - that by default is connected to the
// host PC via the USP programming cable - to print debug messages to the
// serial terminal inside CubeIDE (could be a separate TErmite terminal too);
// Note that debug or other information could be printed on the LCD display too;
char tx_buff_to_HostPC[BUFF_SIZE];
char rx_buff_from_HostPC[BUFF_SIZE];

///////////////////////////////////////////////////////////////////////////////
//
// main program
//
///////////////////////////////////////////////////////////////////////////////
int main(void)
{
  // Reset of all peripherals, Initializes the Flash interface and the Systick.
  HAL_Init();
  // Configure the system clock
  SystemClock_Config();
  // Initialize all configured peripherals
  MX_GPIO_Init();
  MX_USART2_UART_Init();
  MX_SPI1_Init();
  MX_USART1_UART_Init();

  // LCD display initialization; black screen;
  LCD_init();
  UG_FillScreen(C_BLUE);
  LCD_PutStr(50, 56, "STARTING BT EXAMPLE 1", DEFAULT_FONT, C_GREEN, C_BLACK);

  // rename the BLueTooth (HC-05 or HC-06) module;
  // Note1: you should run this function on the Nucleo board only one time in order to
  // rename your own BT module with an unique name; once that is done you should comment
  // this out so that you do not rename the BT module every time;
  // Note2: when running this function it is possible to see nothing returned
  // by BT module and the name wof module would not be changed either; in that case,
  // press the push-button on the BT board and then run this program again;
  //BT_module_rename();

  while (1)
  {
    // Uncomment only one of the following examples, while others are commented out:

    // (1) Example 1:
    // use BT module to just send a hard-codded message out; so, communication is
    // only outwards, broadcasting; we receive this broadcast message with a
    // BlueTooth-to-Serial terminal app on a smartphone (you should search and
    // download such an app, there are many on google play or apple market);
    // send message via BT module connected to UART1;
    HAL_UART_Transmit(&huart1, (uint8_t*)tx_buff_to_BlueTooth_module, strlen(tx_buff_to_BlueTooth_module), HAL_MAX_DELAY);
    // send also a message to host PC via UART2; we can place the message we want to
    // send into the tx_buff_to_HostPC using sprintf or strcpy;
    sprintf(tx_buff_to_HostPC, "%s", "Now sending message via BT module...\r\n");
    HAL_UART_Transmit(&huart2, (uint8_t*)tx_buff_to_HostPC, strlen(tx_buff_to_HostPC), HAL_MAX_DELAY);
    HAL_Delay( 5000);

    // (2) Example 2:
    // TODO as assignment;

    // (3) Example 3:
    // TODO as assignment;

  } // while(1) loop;

}

///////////////////////////////////////////////////////////////////////////////
//
// comm. with host PC via UART2 related functions
//
///////////////////////////////////////////////////////////////////////////////

void readFromHostPC( uint32_t timeout)
{
  // reads from host PC;
  int i = 0;

  // (1) clear receive buffer first;
  for (i = 0; i < BUFF_SIZE; i++) {
    rx_buff_from_HostPC[i] = 0;
  }

  // (2) check if anything was sent from host PC;
  // place received message rx_buff_from_HostPC;
  HAL_UART_Receive(&huart2, (uint8_t *)rx_buff_from_HostPC, BUFF_SIZE, timeout);
}

///////////////////////////////////////////////////////////////////////////////
//
// BT module related functions
//
///////////////////////////////////////////////////////////////////////////////

void readFromBT( uint32_t timeout, int debug)
{
  // reads from the BT module if it has received anything;
  int i = 0;

  // (1) clear receive buffer first;
  for (i = 0; i < BUFF_SIZE; i++) {
    rx_buff_from_BlueTooth_module[i] = 0;
  }

  // (2) check if anything was sent from BT module;
  // place received message into rx_buff_from_BlueTooth_module;
  HAL_UART_Receive(&huart1, (uint8_t *)rx_buff_from_BlueTooth_module, BUFF_SIZE, timeout);

  // (3) if debug is true, we also print to host PC;
  if (debug) {
    sprintf(tx_buff_to_HostPC, "%s", "\r\n<-------- START received data from BT -------->\r\n");
    HAL_UART_Transmit(&huart2, (uint8_t*)tx_buff_to_HostPC, strlen(tx_buff_to_HostPC), HAL_MAX_DELAY);
    strcpy(tx_buff_to_HostPC, rx_buff_from_BlueTooth_module);
    HAL_UART_Transmit(&huart2, (uint8_t*)tx_buff_to_HostPC, strlen(tx_buff_to_HostPC), HAL_MAX_DELAY);
    sprintf(tx_buff_to_HostPC, "%s", "\r\n<-------- END received data from BT -------->\r\n");
    HAL_UART_Transmit(&huart2, (uint8_t*)tx_buff_to_HostPC, strlen(tx_buff_to_HostPC), HAL_MAX_DELAY);
  }
}

void sendCommandToBT(char *command, uint32_t timeout, int debug)
{
  int i = 0;
  // (1) clear receive buffer first;
  for (i = 0; i < BUFF_SIZE; i++) {
    rx_buff_from_BlueTooth_module[i] = 0;
  }

  // (2) send command to BT module;
  HAL_UART_Transmit(&huart1, (uint8_t *)command, strlen(command), timeout);

  // (3) check if BT module replied with anything; place received message
  // into rx_buff_from_BlueTooth_module;
  HAL_UART_Receive(&huart1, (uint8_t *)rx_buff_from_BlueTooth_module, BUFF_SIZE, timeout);

  // (4) if debug is true, we also print to host PC;
  if (debug) {
    sprintf(tx_buff_to_HostPC, "%s", "\r\n<-------- START response to sendCommandToBT() -------->\r\n");
    HAL_UART_Transmit(&huart2, (uint8_t*)tx_buff_to_HostPC, strlen(tx_buff_to_HostPC), HAL_MAX_DELAY);
    strcpy(tx_buff_to_HostPC, rx_buff_from_BlueTooth_module);
    HAL_UART_Transmit(&huart2, (uint8_t*)tx_buff_to_HostPC, strlen(tx_buff_to_HostPC), HAL_MAX_DELAY);
    sprintf(tx_buff_to_HostPC, "%s", "\r\n<-------- END response to sendCommandToBT() -------->\r\n");
    HAL_UART_Transmit(&huart2, (uint8_t*)tx_buff_to_HostPC, strlen(tx_buff_to_HostPC), HAL_MAX_DELAY);
  }
}


void BT_module_rename(void)
{
  // this function should be called only once;
  sendCommandToBT("AT\r\n", 3000, 1); // timeout of 3000 ms should be enough; debug set to true;
  HAL_Delay(1000);
  sendCommandToBT("AT+VERSION\r\n", 3000, 1); // ask version of firmware;
  HAL_Delay(1000);
  sendCommandToBT("AT+NAME=CRIS-NUCLEO\r\n", 3000, 1); // change name of BT module to CRIS-NUCLEO
  HAL_Delay(1000);
}


///////////////////////////////////////////////////////////////////////////////
//
// configuration functions
//
///////////////////////////////////////////////////////////////////////////////

/**
  * @brief System Clock Configuration
  * @retval None
  */
void SystemClock_Config(void)
{
  RCC_OscInitTypeDef RCC_OscInitStruct = {0};
  RCC_ClkInitTypeDef RCC_ClkInitStruct = {0};
  RCC_PeriphCLKInitTypeDef PeriphClkInit = {0};

  /** Configure the main internal regulator output voltage
  */
  __HAL_PWR_VOLTAGESCALING_CONFIG(PWR_REGULATOR_VOLTAGE_SCALE1);

  /** Initializes the RCC Oscillators according to the specified parameters
  * in the RCC_OscInitTypeDef structure.
  */
  RCC_OscInitStruct.OscillatorType = RCC_OSCILLATORTYPE_HSI;
  RCC_OscInitStruct.HSIState = RCC_HSI_ON;
  RCC_OscInitStruct.HSICalibrationValue = RCC_HSICALIBRATION_DEFAULT;
  RCC_OscInitStruct.PLL.PLLState = RCC_PLL_NONE;
  if (HAL_RCC_OscConfig(&RCC_OscInitStruct) != HAL_OK)
  {
    Error_Handler();
  }

  /** Initializes the CPU, AHB and APB buses clocks
  */
  RCC_ClkInitStruct.ClockType = RCC_CLOCKTYPE_HCLK|RCC_CLOCKTYPE_SYSCLK
                              |RCC_CLOCKTYPE_PCLK1|RCC_CLOCKTYPE_PCLK2;
  RCC_ClkInitStruct.SYSCLKSource = RCC_SYSCLKSOURCE_HSI;
  RCC_ClkInitStruct.AHBCLKDivider = RCC_SYSCLK_DIV1;
  RCC_ClkInitStruct.APB1CLKDivider = RCC_HCLK_DIV1;
  RCC_ClkInitStruct.APB2CLKDivider = RCC_HCLK_DIV1;

  if (HAL_RCC_ClockConfig(&RCC_ClkInitStruct, FLASH_LATENCY_0) != HAL_OK)
  {
    Error_Handler();
  }
  PeriphClkInit.PeriphClockSelection = RCC_PERIPHCLK_USART1|RCC_PERIPHCLK_USART2;
  PeriphClkInit.Usart1ClockSelection = RCC_USART1CLKSOURCE_PCLK2;
  PeriphClkInit.Usart2ClockSelection = RCC_USART2CLKSOURCE_PCLK1;
  if (HAL_RCCEx_PeriphCLKConfig(&PeriphClkInit) != HAL_OK)
  {
    Error_Handler();
  }
}

/**
  * @brief SPI1 Initialization Function
  * @param None
  * @retval None
  */
static void MX_SPI1_Init(void)
{

  /* USER CODE BEGIN SPI1_Init 0 */

  /* USER CODE END SPI1_Init 0 */

  /* USER CODE BEGIN SPI1_Init 1 */

  /* USER CODE END SPI1_Init 1 */
  /* SPI1 parameter configuration*/
  hspi1.Instance = SPI1;
  hspi1.Init.Mode = SPI_MODE_MASTER;
  hspi1.Init.Direction = SPI_DIRECTION_2LINES;
  hspi1.Init.DataSize = SPI_DATASIZE_8BIT;
  hspi1.Init.CLKPolarity = SPI_POLARITY_HIGH;
  hspi1.Init.CLKPhase = SPI_PHASE_1EDGE;
  hspi1.Init.NSS = SPI_NSS_SOFT;
  hspi1.Init.BaudRatePrescaler = SPI_BAUDRATEPRESCALER_4;
  hspi1.Init.FirstBit = SPI_FIRSTBIT_MSB;
  hspi1.Init.TIMode = SPI_TIMODE_DISABLE;
  hspi1.Init.CRCCalculation = SPI_CRCCALCULATION_DISABLE;
  hspi1.Init.CRCPolynomial = 7;
  if (HAL_SPI_Init(&hspi1) != HAL_OK)
  {
    Error_Handler();
  }
  /* USER CODE BEGIN SPI1_Init 2 */

  /* USER CODE END SPI1_Init 2 */

}

/**
  * @brief USART1 Initialization Function
  * @param None
  * @retval None
  */
static void MX_USART1_UART_Init(void)
{

  /* USER CODE BEGIN USART1_Init 0 */

  /* USER CODE END USART1_Init 0 */

  /* USER CODE BEGIN USART1_Init 1 */

  /* USER CODE END USART1_Init 1 */
  huart1.Instance = USART1;
  huart1.Init.BaudRate = 9600;
  huart1.Init.WordLength = UART_WORDLENGTH_8B;
  huart1.Init.StopBits = UART_STOPBITS_1;
  huart1.Init.Parity = UART_PARITY_NONE;
  huart1.Init.Mode = UART_MODE_TX_RX;
  huart1.Init.HwFlowCtl = UART_HWCONTROL_NONE;
  huart1.Init.OverSampling = UART_OVERSAMPLING_16;
  huart1.Init.OneBitSampling = UART_ONE_BIT_SAMPLE_DISABLE;
  huart1.AdvancedInit.AdvFeatureInit = UART_ADVFEATURE_NO_INIT;
  if (HAL_UART_Init(&huart1) != HAL_OK)
  {
    Error_Handler();
  }
  /* USER CODE BEGIN USART1_Init 2 */

  /* USER CODE END USART1_Init 2 */

}

/**
  * @brief USART2 Initialization Function
  * @param None
  * @retval None
  */
static void MX_USART2_UART_Init(void)
{

  /* USER CODE BEGIN USART2_Init 0 */

  /* USER CODE END USART2_Init 0 */

  /* USER CODE BEGIN USART2_Init 1 */

  /* USER CODE END USART2_Init 1 */
  huart2.Instance = USART2;
  huart2.Init.BaudRate = 115200;
  huart2.Init.WordLength = UART_WORDLENGTH_8B;
  huart2.Init.StopBits = UART_STOPBITS_1;
  huart2.Init.Parity = UART_PARITY_NONE;
  huart2.Init.Mode = UART_MODE_TX_RX;
  huart2.Init.HwFlowCtl = UART_HWCONTROL_NONE;
  huart2.Init.OverSampling = UART_OVERSAMPLING_16;
  huart2.Init.OneBitSampling = UART_ONE_BIT_SAMPLE_DISABLE;
  huart2.AdvancedInit.AdvFeatureInit = UART_ADVFEATURE_NO_INIT;
  if (HAL_UART_Init(&huart2) != HAL_OK)
  {
    Error_Handler();
  }
  /* USER CODE BEGIN USART2_Init 2 */

  /* USER CODE END USART2_Init 2 */

}

/**
  * @brief GPIO Initialization Function
  * @param None
  * @retval None
  */
static void MX_GPIO_Init(void)
{
  GPIO_InitTypeDef GPIO_InitStruct = {0};
/* USER CODE BEGIN MX_GPIO_Init_1 */
/* USER CODE END MX_GPIO_Init_1 */

  /* GPIO Ports Clock Enable */
  __HAL_RCC_GPIOC_CLK_ENABLE();
  __HAL_RCC_GPIOH_CLK_ENABLE();
  __HAL_RCC_GPIOA_CLK_ENABLE();
  __HAL_RCC_GPIOB_CLK_ENABLE();

  /*Configure GPIO pin Output Level */
  HAL_GPIO_WritePin(GPIOA, LD2_Pin|LCD_D_C_Pin, GPIO_PIN_RESET);

  /*Configure GPIO pin Output Level */
  HAL_GPIO_WritePin(LCD_RST_GPIO_Port, LCD_RST_Pin, GPIO_PIN_RESET);

  /*Configure GPIO pin Output Level */
  HAL_GPIO_WritePin(LCD_CS_GPIO_Port, LCD_CS_Pin, GPIO_PIN_RESET);

  /*Configure GPIO pin : B1_Pin */
  GPIO_InitStruct.Pin = B1_Pin;
  GPIO_InitStruct.Mode = GPIO_MODE_IT_FALLING;
  GPIO_InitStruct.Pull = GPIO_NOPULL;
  HAL_GPIO_Init(B1_GPIO_Port, &GPIO_InitStruct);

  /*Configure GPIO pins : LD2_Pin LCD_D_C_Pin */
  GPIO_InitStruct.Pin = LD2_Pin|LCD_D_C_Pin;
  GPIO_InitStruct.Mode = GPIO_MODE_OUTPUT_PP;
  GPIO_InitStruct.Pull = GPIO_NOPULL;
  GPIO_InitStruct.Speed = GPIO_SPEED_FREQ_LOW;
  HAL_GPIO_Init(GPIOA, &GPIO_InitStruct);

  /*Configure GPIO pin : LCD_RST_Pin */
  GPIO_InitStruct.Pin = LCD_RST_Pin;
  GPIO_InitStruct.Mode = GPIO_MODE_OUTPUT_PP;
  GPIO_InitStruct.Pull = GPIO_NOPULL;
  GPIO_InitStruct.Speed = GPIO_SPEED_FREQ_LOW;
  HAL_GPIO_Init(LCD_RST_GPIO_Port, &GPIO_InitStruct);

  /*Configure GPIO pin : LCD_CS_Pin */
  GPIO_InitStruct.Pin = LCD_CS_Pin;
  GPIO_InitStruct.Mode = GPIO_MODE_OUTPUT_PP;
  GPIO_InitStruct.Pull = GPIO_NOPULL;
  GPIO_InitStruct.Speed = GPIO_SPEED_FREQ_LOW;
  HAL_GPIO_Init(LCD_CS_GPIO_Port, &GPIO_InitStruct);

/* USER CODE BEGIN MX_GPIO_Init_2 */
/* USER CODE END MX_GPIO_Init_2 */
}

/* USER CODE BEGIN 4 */

/* USER CODE END 4 */

/**
  * @brief  This function is executed in case of error occurrence.
  * @retval None
  */
void Error_Handler(void)
{
  /* USER CODE BEGIN Error_Handler_Debug */
  /* User can add his own implementation to report the HAL error return state */
  __disable_irq();
  while (1)
  {
  }
  /* USER CODE END Error_Handler_Debug */
}

#ifdef  USE_FULL_ASSERT
/**
  * @brief  Reports the name of the source file and the source line number
  *         where the assert_param error has occurred.
  * @param  file: pointer to the source file name
  * @param  line: assert_param error line source number
  * @retval None
  */
void assert_failed(uint8_t *file, uint32_t line)
{
  /* USER CODE BEGIN 6 */
  /* User can add his own implementation to report the file name and line number,
     ex: printf("Wrong parameters value: file %s on line %d\r\n", file, line) */
  /* USER CODE END 6 */
}
#endif /* USE_FULL_ASSERT */
